@@ -17,14 +17,17 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
-
-#include "glyphbox.h"
-#include "extract.h"
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
+#if FZ_ENABLE_DOCX_OUTPUT
+
+#include "glyphbox.h"
+#include "extract/extract.h"
+#include "extract/buffer.h"
+
 #include <assert.h>
 #include <errno.h>
 #include <string.h>
@@ -93,7 +96,7 @@ static void dev_text(fz_context *ctx, fz_device *dev_, const fz_text *text, fz_m
 					bbox.x1,
 					bbox.y1))
 			{
-				fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to begin span");
+				fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to begin span");
 			}
 
 			trm = span->trm;
@@ -118,11 +121,11 @@ static void dev_text(fz_context *ctx, fz_device *dev_, const fz_text *text, fz_m
 				bounds = fz_bound_glyph(ctx, span->font, span->items[i].gid, combined);
 				if (extract_add_char(dev->writer->extract, combined.e, combined.f, item->ucs, adv,
 							bounds.x0, bounds.y0, bounds.x1, bounds.y1))
-					fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to add char");
+					fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to add char");
 			}
 
 			if (extract_span_end(dev->writer->extract))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to end span");
+				fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to end span");
 		}
 	}
 	fz_always(ctx)
@@ -213,7 +216,7 @@ static void dev_fill_image(fz_context *ctx, fz_device *dev_, fz_image *img, fz_m
 						dev->writer
 						))
 				{
-					fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to add image type=%s", type);
+					fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to add image type=%s", type);
 				}
 			}
 			else
@@ -253,14 +256,14 @@ static void s_moveto(fz_context *ctx, void *arg, float x, float y)
 {
 	extract_t* extract = arg;
 	if (extract_moveto(extract, x, y))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "extract_moveto() failed");
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_moveto() failed");
 }
 
 static void s_lineto(fz_context *ctx, void *arg, float x, float y)
 {
 	extract_t* extract = arg;
 	if (extract_lineto(extract, x, y))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "extract_lineto() failed");
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_lineto() failed");
 }
 
 static void s_curveto(fz_context *ctx, void *arg, float x1, float y1,
@@ -270,14 +273,14 @@ static void s_curveto(fz_context *ctx, void *arg, float x1, float y1,
 	(straight) lines will be handled correctly. */
 	extract_t* extract = arg;
 	if (extract_moveto(extract, x3, y3))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "extract_moveto() failed");
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_moveto() failed");
 }
 
 static void s_closepath(fz_context *ctx, void *arg)
 {
 	extract_t* extract = arg;
 	if (extract_closepath(extract))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "extract_closepath() failed");
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_closepath() failed");
 }
 
 /*
@@ -322,10 +325,10 @@ void dev_fill_path(fz_context *ctx, fz_device *dev_, const fz_path *path, int ev
 				matrix.f,
 				color[0]
 				))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to begin fill");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to begin fill");
 		s_walk_path(ctx, dev, extract, path);
 		if (extract_fill_end(extract))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "extract_fill_end() failed");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_fill_end() failed");
 	}
 	fz_always(ctx)
 	{
@@ -362,10 +365,10 @@ dev_stroke_path(fz_context *ctx, fz_device *dev_, const fz_path *path,
 				stroke->linewidth,
 				color[0]
 				))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to begin stroke");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to begin stroke");
 		s_walk_path(ctx, dev, extract, path);
 		if (extract_stroke_end(extract))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "extract_stroke_end() failed");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_stroke_end() failed");
 	}
 	fz_always(ctx)
 	{
@@ -377,6 +380,191 @@ dev_stroke_path(fz_context *ctx, fz_device *dev_, const fz_path *path,
 	}
 }
 
+static extract_struct_t
+fz_struct_to_extract(fz_structure type)
+{
+	switch (type)
+	{
+	default:
+		return extract_struct_INVALID;
+
+	case FZ_STRUCTURE_DOCUMENT:
+		return extract_struct_DOCUMENT;
+	case FZ_STRUCTURE_PART:
+		return extract_struct_PART;
+	case FZ_STRUCTURE_ART:
+		return extract_struct_ART;
+	case FZ_STRUCTURE_SECT:
+		return extract_struct_SECT;
+	case FZ_STRUCTURE_DIV:
+		return extract_struct_DIV;
+	case FZ_STRUCTURE_BLOCKQUOTE:
+		return extract_struct_BLOCKQUOTE;
+	case FZ_STRUCTURE_CAPTION:
+		return extract_struct_CAPTION;
+	case FZ_STRUCTURE_TOC:
+		return extract_struct_TOC;
+	case FZ_STRUCTURE_TOCI:
+		return extract_struct_TOCI;
+	case FZ_STRUCTURE_INDEX:
+		return extract_struct_INDEX;
+	case FZ_STRUCTURE_NONSTRUCT:
+		return extract_struct_NONSTRUCT;
+	case FZ_STRUCTURE_PRIVATE:
+		return extract_struct_PRIVATE;
+	/* Grouping elements (PDF 2.0 - Table 364) */
+	case FZ_STRUCTURE_DOCUMENTFRAGMENT:
+		return extract_struct_DOCUMENTFRAGMENT;
+	/* Grouping elements (PDF 2.0 - Table 365) */
+	case FZ_STRUCTURE_ASIDE:
+		return extract_struct_ASIDE;
+	/* Grouping elements (PDF 2.0 - Table 366) */
+	case FZ_STRUCTURE_TITLE:
+		return extract_struct_TITLE;
+	case FZ_STRUCTURE_FENOTE:
+		return extract_struct_FENOTE;
+	/* Grouping elements (PDF 2.0 - Table 367) */
+	case FZ_STRUCTURE_SUB:
+		return extract_struct_SUB;
+
+	/* Paragraphlike elements (PDF 1.7 - Table 10.21) */
+	case FZ_STRUCTURE_P:
+		return extract_struct_P;
+	case FZ_STRUCTURE_H:
+		return extract_struct_H;
+	case FZ_STRUCTURE_H1:
+		return extract_struct_H1;
+	case FZ_STRUCTURE_H2:
+		return extract_struct_H2;
+	case FZ_STRUCTURE_H3:
+		return extract_struct_H3;
+	case FZ_STRUCTURE_H4:
+		return extract_struct_H4;
+	case FZ_STRUCTURE_H5:
+		return extract_struct_H5;
+	case FZ_STRUCTURE_H6:
+		return extract_struct_H6;
+
+	/* List elements (PDF 1.7 - Table 10.23) */
+	case FZ_STRUCTURE_LIST:
+		return extract_struct_LIST;
+	case FZ_STRUCTURE_LISTITEM:
+		return extract_struct_LISTITEM;
+	case FZ_STRUCTURE_LABEL:
+		return extract_struct_LABEL;
+	case FZ_STRUCTURE_LISTBODY:
+		return extract_struct_LISTBODY;
+
+	/* Table elements (PDF 1.7 - Table 10.24) */
+	case FZ_STRUCTURE_TABLE:
+		return extract_struct_TABLE;
+	case FZ_STRUCTURE_TR:
+		return extract_struct_TR;
+	case FZ_STRUCTURE_TH:
+		return extract_struct_TH;
+	case FZ_STRUCTURE_TD:
+		return extract_struct_TD;
+	case FZ_STRUCTURE_THEAD:
+		return extract_struct_THEAD;
+	case FZ_STRUCTURE_TBODY:
+		return extract_struct_TBODY;
+	case FZ_STRUCTURE_TFOOT:
+		return extract_struct_TFOOT;
+
+	/* Inline elements (PDF 1.7 - Table 10.25) */
+	case FZ_STRUCTURE_SPAN:
+		return extract_struct_SPAN;
+	case FZ_STRUCTURE_QUOTE:
+		return extract_struct_QUOTE;
+	case FZ_STRUCTURE_NOTE:
+		return extract_struct_NOTE;
+	case FZ_STRUCTURE_REFERENCE:
+		return extract_struct_REFERENCE;
+	case FZ_STRUCTURE_BIBENTRY:
+		return extract_struct_BIBENTRY;
+	case FZ_STRUCTURE_CODE:
+		return extract_struct_CODE;
+	case FZ_STRUCTURE_LINK:
+		return extract_struct_LINK;
+	case FZ_STRUCTURE_ANNOT:
+		return extract_struct_ANNOT;
+	/* Inline elements (PDF 2.0 - Table 368) */
+	case FZ_STRUCTURE_EM:
+		return extract_struct_EM;
+	case FZ_STRUCTURE_STRONG:
+		return extract_struct_STRONG;
+
+	/* Ruby inline element (PDF 1.7 - Table 10.26) */
+	case FZ_STRUCTURE_RUBY:
+		return extract_struct_RUBY;
+	case FZ_STRUCTURE_RB:
+		return extract_struct_RB;
+	case FZ_STRUCTURE_RT:
+		return extract_struct_RT;
+	case FZ_STRUCTURE_RP:
+		return extract_struct_RP;
+
+	/* Warichu inline element (PDF 1.7 - Table 10.26) */
+	case FZ_STRUCTURE_WARICHU:
+		return extract_struct_WARICHU;
+	case FZ_STRUCTURE_WT:
+		return extract_struct_WT;
+	case FZ_STRUCTURE_WP:
+		return extract_struct_WP;
+
+	/* Illustration elements (PDF 1.7 - Table 10.27) */
+	case FZ_STRUCTURE_FIGURE:
+		return extract_struct_FIGURE;
+	case FZ_STRUCTURE_FORMULA:
+		return extract_struct_FORMULA;
+	case FZ_STRUCTURE_FORM:
+		return extract_struct_FORM;
+
+	/* Artifact structure type (PDF 2.0 - Table 375) */
+	case FZ_STRUCTURE_ARTIFACT:
+		return extract_struct_ARTIFACT;
+	}
+}
+
+static void
+dev_begin_structure(fz_context *ctx, fz_device *dev_, fz_structure standard, const char *raw, int idx)
+{
+	fz_docx_device *dev = (fz_docx_device *)dev_;
+	extract_t *extract = dev->writer->extract;
+
+	assert(!dev->writer->ctx);
+	dev->writer->ctx = ctx;
+	fz_try(ctx)
+	{
+		if (extract_begin_struct(extract, fz_struct_to_extract(standard), idx, -1))
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to begin struct");
+	}
+	fz_always(ctx)
+		dev->writer->ctx = NULL;
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
+static void
+dev_end_structure(fz_context *ctx, fz_device *dev_)
+{
+	fz_docx_device *dev = (fz_docx_device *)dev_;
+	extract_t *extract = dev->writer->extract;
+
+	assert(!dev->writer->ctx);
+	dev->writer->ctx = ctx;
+	fz_try(ctx)
+	{
+		if (extract_end_struct(extract))
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to end struct");
+	}
+	fz_always(ctx)
+		dev->writer->ctx = NULL;
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
+
 static fz_device *writer_begin_page(fz_context *ctx, fz_document_writer *writer_, fz_rect mediabox)
 {
 	fz_docx_writer *writer = (fz_docx_writer*) writer_;
@@ -388,7 +576,7 @@ static fz_device *writer_begin_page(fz_context *ctx, fz_document_writer *writer_
 	fz_try(ctx)
 	{
 		if (extract_page_begin(writer->extract, mediabox.x0, mediabox.y0, mediabox.x1, mediabox.y1))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to begin page");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to begin page");
 		dev = fz_new_derived_device(ctx, fz_docx_device);
 		dev->super.fill_text = dev_fill_text;
 		dev->super.stroke_text = dev_stroke_text;
@@ -398,6 +586,8 @@ static fz_device *writer_begin_page(fz_context *ctx, fz_document_writer *writer_
 		dev->super.fill_image = dev_fill_image;
 		dev->super.fill_path = dev_fill_path;
 		dev->super.stroke_path = dev_stroke_path;
+		dev->super.begin_structure = dev_begin_structure;
+		dev->super.end_structure = dev_end_structure;
 		dev->writer = writer;
 	}
 	fz_always(ctx)
@@ -420,10 +610,10 @@ static void writer_end_page(fz_context *ctx, fz_document_writer *writer_, fz_dev
 	{
 		fz_close_device(ctx, dev);
 		if (extract_page_end(writer->extract))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to end page");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to end page");
 
 		if (extract_process(writer->extract, writer->spacing, writer->rotation, writer->images))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to process page");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to process page");
 	}
 	fz_always(ctx)
 	{
@@ -495,12 +685,12 @@ static void writer_close(fz_context *ctx, fz_document_writer *writer_)
 				&extract_buffer_output
 				))
 		{
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to create extract_buffer_output: %s", strerror(errno));
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to create extract_buffer_output: %s", strerror(errno));
 		}
 		if (extract_write(writer->extract, extract_buffer_output))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to generate docx content: %s", strerror(errno));
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to generate docx content: %s", strerror(errno));
 		if (extract_buffer_close(&extract_buffer_output))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to close extract_buffer: %s", strerror(errno));
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to close extract_buffer: %s", strerror(errno));
 
 		extract_end(&writer->extract);
 		fz_close_output(ctx, writer->output);
@@ -546,6 +736,18 @@ static int get_bool_option(fz_context *ctx, const char *options, const char *nam
 		return default_;
 }
 
+static double get_double_option(fz_context *ctx, const char *options, const char *name, double default_)
+{
+	const char *value;
+	if (fz_has_option(ctx, options, name, &value))
+	{
+		double ret = atof(value);
+		return ret;
+	}
+	else
+		return default_;
+}
+
 static void *s_realloc_fn(void *state, void *prev, size_t size)
 {
 	fz_docx_writer *writer = state;
@@ -564,6 +766,7 @@ static fz_document_writer *fz_new_docx_writer_internal(fz_context *ctx, fz_outpu
 
 	fz_try(ctx)
 	{
+		double space_guess = get_double_option(ctx, options, "space-guess", 0);
 		writer = fz_new_derived_document_writer(
 				ctx,
 				fz_docx_writer,
@@ -576,16 +779,19 @@ static fz_document_writer *fz_new_docx_writer_internal(fz_context *ctx, fz_outpu
 		writer->output = out;
 		if (get_bool_option(ctx, options, "html", 0)) format = extract_format_HTML;
 		if (get_bool_option(ctx, options, "text", 0)) format = extract_format_TEXT;
+		if (get_bool_option(ctx, options, "json", 0)) format = extract_format_JSON;
 		if (extract_alloc_create(s_realloc_fn, writer, &writer->alloc))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to create extract_alloc instance");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to create extract_alloc instance");
 		if (extract_begin(writer->alloc, format, &writer->extract))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Failed to create extract instance");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "Failed to create extract instance");
+		if (space_guess)
+			extract_set_space_guess(writer->extract, space_guess);
 		writer->spacing = get_bool_option(ctx, options, "spacing", 0);
 		writer->rotation = get_bool_option(ctx, options, "rotation", 1);
 		writer->images = get_bool_option(ctx, options, "images", 1);
 		writer->mediabox_clip = get_bool_option(ctx, options, "mediabox-clip", 1);
 		if (extract_set_layout_analysis(writer->extract, get_bool_option(ctx, options, "analyse", 0)))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "extract_enable_analysis failed.");
+			fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_enable_analysis failed.");
 		{
 			const char* v;
 			if (fz_has_option(ctx, options, "tables-csv-format", &v))
@@ -597,7 +803,7 @@ static fz_document_writer *fz_new_docx_writer_internal(fz_context *ctx, fz_outpu
 				if (extract_tables_csv_format(writer->extract, formatbuf))
 				{
 					fz_free(ctx, formatbuf);
-					fz_throw(ctx, FZ_ERROR_GENERIC, "extract_tables_csv_format() failed.");
+					fz_throw(ctx, FZ_ERROR_LIBRARY, "extract_tables_csv_format() failed.");
 				}
 				fz_free(ctx, formatbuf);
 			}
@@ -621,16 +827,26 @@ static fz_document_writer *fz_new_docx_writer_internal(fz_context *ctx, fz_outpu
 	return &writer->super;
 }
 
-fz_document_writer *fz_new_odt_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
-{
-	return fz_new_docx_writer_internal(ctx, out, options, extract_format_ODT);
-}
-
 fz_document_writer *fz_new_docx_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
 {
 	return fz_new_docx_writer_internal(ctx, out, options, extract_format_DOCX);
 }
 
+fz_document_writer *fz_new_docx_writer(fz_context *ctx, const char *path, const char *options)
+{
+	/* No need to drop <out> if fz_new_docx_writer_internal() throws, because
+	it always drops <out> if it fails. */
+	fz_output *out = fz_new_output_with_path(ctx, path, 0 /*append*/);
+	return fz_new_docx_writer_internal(ctx, out, options, extract_format_DOCX);
+}
+
+#if FZ_ENABLE_ODT_OUTPUT
+
+fz_document_writer *fz_new_odt_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
+{
+	return fz_new_docx_writer_internal(ctx, out, options, extract_format_ODT);
+}
+
 fz_document_writer *fz_new_odt_writer(fz_context *ctx, const char *path, const char *options)
 {
 	/* No need to drop <out> if fz_new_docx_writer_internal() throws, because
@@ -639,10 +855,46 @@ fz_document_writer *fz_new_odt_writer(fz_context *ctx, const char *path, const c
 	return fz_new_docx_writer_internal(ctx, out, options, extract_format_ODT);
 }
 
+#else
+
+fz_document_writer *fz_new_odt_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
+{
+	fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "ODT writer not enabled");
+	return NULL;
+}
+
+fz_document_writer *fz_new_odt_writer(fz_context *ctx, const char *path, const char *options)
+{
+	fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "ODT writer not enabled");
+	return NULL;
+}
+
+#endif
+
+#else
+
+fz_document_writer *fz_new_odt_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
+{
+	fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "DOCX/ODT writer not enabled");
+	return NULL;
+}
+
+fz_document_writer *fz_new_odt_writer(fz_context *ctx, const char *path, const char *options)
+{
+	fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "DOCX/ODT writer not enabled");
+	return NULL;
+}
+
+fz_document_writer *fz_new_docx_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
+{
+	fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "DOCX writer not enabled");
+	return NULL;
+}
+
 fz_document_writer *fz_new_docx_writer(fz_context *ctx, const char *path, const char *options)
 {
-	/* No need to drop <out> if fz_new_docx_writer_internal() throws, because
-	it always drops <out> if it fails. */
-	fz_output *out = fz_new_output_with_path(ctx, path, 0 /*append*/);
-	return fz_new_docx_writer_internal(ctx, out, options, extract_format_DOCX);
+	fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "DOCX writer not enabled");
+	return NULL;
 }
+
+#endif
