@@ -33,6 +33,8 @@ int
 pdf_count_pages(fz_context *ctx, pdf_document *doc)
 {
 	int pages;
+	if (doc->is_fdf)
+		return 0;
 	/* FIXME: We should reset linear_page_count to 0 when editing starts
 	 * (or when linear loading ends) */
 	if (doc->linear_page_count != 0)
@@ -40,7 +42,7 @@ pdf_count_pages(fz_context *ctx, pdf_document *doc)
 	else
 		pages = pdf_to_int(ctx, pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/Pages/Count"));
 	if (pages < 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Invalid number of pages");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "Invalid number of pages");
 	return pages;
 }
 
@@ -59,14 +61,14 @@ pdf_load_page_tree_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int id
 		pdf_obj *kids = pdf_dict_get(ctx, node, PDF_NAME(Kids));
 		int i, n = pdf_array_len(ctx, kids);
 		if (pdf_cycle(ctx, &cycle, cycle_up, node))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in page tree");
 		for (i = 0; i < n; ++i)
 			idx = pdf_load_page_tree_imp(ctx, doc, pdf_array_get(ctx, kids, i), idx, &cycle);
 	}
 	else if (pdf_name_eq(ctx, type, PDF_NAME(Page)))
 	{
 		if (idx >= doc->map_page_count)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "too many kids in page tree");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "too many kids in page tree");
 		doc->rev_page_map[idx].page = idx;
 		doc->rev_page_map[idx].object = pdf_to_num(ctx, node);
 		doc->fwd_page_map[idx] = pdf_keep_obj(ctx, node);
@@ -74,7 +76,7 @@ pdf_load_page_tree_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int id
 	}
 	else
 	{
-		fz_throw(ctx, FZ_ERROR_GENERIC, "non-page object in page tree");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "non-page object in page tree");
 	}
 	return idx;
 }
@@ -118,10 +120,25 @@ pdf_load_page_tree_internal(fz_context *ctx, pdf_document *doc)
 	 * stuff that hits the document at a time. */
 	fz_try(ctx)
 	{
+		int idx;
+
 		doc->map_page_count = pdf_count_pages(ctx, doc);
-		doc->rev_page_map = Memento_label(fz_calloc(ctx, doc->map_page_count, sizeof(pdf_rev_page_map)), "pdf_rev_page_map");
-		doc->fwd_page_map = Memento_label(fz_calloc(ctx, doc->map_page_count, sizeof(pdf_obj *)), "pdf_fwd_page_map");
-		pdf_load_page_tree_imp(ctx, doc, pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/Pages"), 0, NULL);
+		while (1)
+		{
+			doc->rev_page_map = Memento_label(fz_calloc(ctx, doc->map_page_count, sizeof(pdf_rev_page_map)), "pdf_rev_page_map");
+			doc->fwd_page_map = Memento_label(fz_calloc(ctx, doc->map_page_count, sizeof(pdf_obj *)), "pdf_fwd_page_map");
+			idx = pdf_load_page_tree_imp(ctx, doc, pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/Pages"), 0, NULL);
+			if (idx < doc->map_page_count)
+			{
+				/* The document claims more pages that it has. Fix that. */
+				fz_warn(ctx, "Document claims to have %d pages, but only has %d.", doc->map_page_count, idx);
+				/* This put drops the page tree! */
+				pdf_dict_putp_drop(ctx, pdf_trailer(ctx, doc), "Root/Pages/Count", pdf_new_int(ctx, idx));
+				doc->map_page_count = idx;
+				continue;
+			}
+			break;
+		}
 		qsort(doc->rev_page_map, doc->map_page_count, sizeof *doc->rev_page_map, cmp_rev_page_map);
 	}
 	fz_catch(ctx)
@@ -155,10 +172,10 @@ pdf_lookup_page_loc_imp(fz_context *ctx, pdf_document *doc, pdf_obj *node, int *
 			len = pdf_array_len(ctx, kids);
 
 			if (len == 0)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "malformed page tree");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "malformed page tree");
 
 			if (pdf_mark_list_push(ctx, &mark_list, node))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in page tree");
 
 			for (i = 0; i < len; i++)
 			{
@@ -224,11 +241,11 @@ pdf_lookup_page_loc(fz_context *ctx, pdf_document *doc, int needle, pdf_obj **pa
 	pdf_obj *hit;
 
 	if (!node)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find page tree");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot find page tree");
 
 	hit = pdf_lookup_page_loc_imp(ctx, doc, node, &skip, parentp, indexp);
 	if (!hit)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find page %d in page tree", needle+1);
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot find page %d in page tree", needle+1);
 	return hit;
 }
 
@@ -242,6 +259,8 @@ pdf_lookup_page_obj(fz_context *ctx, pdf_document *doc, int needle)
 		fz_catch(ctx)
 		{
 			doc->page_tree_broken = 1;
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			fz_warn(ctx, "Page tree load failed. Falling back to slow lookup");
 		}
 	}
@@ -249,7 +268,7 @@ pdf_lookup_page_obj(fz_context *ctx, pdf_document *doc, int needle)
 	if (doc->fwd_page_map)
 	{
 		if (needle < 0 || needle >= doc->map_page_count)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find page %d in page tree", needle+1);
+			fz_throw(ctx, FZ_ERROR_FORMAT, "cannot find page %d in page tree", needle+1);
 		if (doc->fwd_page_map[needle] != NULL)
 			return doc->fwd_page_map[needle];
 	}
@@ -272,13 +291,13 @@ pdf_count_pages_before_kid(fz_context *ctx, pdf_document *doc, pdf_obj *parent,
 			pdf_obj *count = pdf_dict_get(ctx, kid, PDF_NAME(Count));
 			int n = pdf_to_int(ctx, count);
 			if (!pdf_is_int(ctx, count) || n < 0)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "illegal or missing count in pages tree");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "illegal or missing count in pages tree");
 			total += n;
 		}
 		else
 			total++;
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "kid not found in parent's kids array");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "kid not found in parent's kids array");
 }
 
 static int
@@ -302,7 +321,7 @@ pdf_lookup_page_number_slow(fz_context *ctx, pdf_document *doc, pdf_obj *node)
 		while (pdf_is_dict(ctx, parent))
 		{
 			if (pdf_mark_list_push(ctx, &mark_list, parent))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in page tree (parents)");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in page tree (parents)");
 			total += pdf_count_pages_before_kid(ctx, doc, parent, needle);
 			needle = pdf_to_num(ctx, parent);
 			parent = pdf_dict_get(ctx, parent, PDF_NAME(Parent));
@@ -562,7 +581,7 @@ pdf_page_presentation(fz_context *ctx, pdf_page *page, fz_transition *transition
 
 	obj = pdf_dict_get(ctx, transdict, PDF_NAME(D));
 
-	transition->duration = (obj ? pdf_to_real(ctx, obj) : 1);
+	transition->duration = pdf_to_real_default(ctx, obj, 1);
 
 	transition->vertical = !pdf_name_eq(ctx, pdf_dict_get(ctx, transdict, PDF_NAME(Dm)), PDF_NAME(H));
 	transition->outwards = !pdf_name_eq(ctx, pdf_dict_get(ctx, transdict, PDF_NAME(M)), PDF_NAME(I));
@@ -605,9 +624,40 @@ fz_rect
 pdf_bound_page(fz_context *ctx, pdf_page *page, fz_box_type box)
 {
 	fz_matrix page_ctm;
-	fz_rect mediabox;
-	pdf_page_transform_box(ctx, page, &mediabox, &page_ctm, box);
-	return fz_transform_rect(mediabox, page_ctm);
+	fz_rect rect;
+	pdf_page_transform_box(ctx, page, &rect, &page_ctm, box);
+	return fz_transform_rect(rect, page_ctm);
+}
+
+void
+pdf_set_page_box(fz_context *ctx, pdf_page *page, fz_box_type box, fz_rect rect)
+{
+	fz_matrix page_ctm, inv_page_ctm;
+	fz_rect page_rect;
+	pdf_page_transform_box(ctx, page, NULL, &page_ctm, box);
+	inv_page_ctm = fz_invert_matrix(page_ctm);
+	page_rect = fz_transform_rect(rect, inv_page_ctm);
+
+	switch (box)
+	{
+	case FZ_MEDIA_BOX:
+		pdf_dict_put_rect(ctx, page->obj, PDF_NAME(MediaBox), page_rect);
+		break;
+	case FZ_CROP_BOX:
+		pdf_dict_put_rect(ctx, page->obj, PDF_NAME(CropBox), page_rect);
+		break;
+	case FZ_BLEED_BOX:
+		pdf_dict_put_rect(ctx, page->obj, PDF_NAME(BleedBox), page_rect);
+		break;
+	case FZ_TRIM_BOX:
+		pdf_dict_put_rect(ctx, page->obj, PDF_NAME(TrimBox), page_rect);
+		break;
+	case FZ_ART_BOX:
+		pdf_dict_put_rect(ctx, page->obj, PDF_NAME(ArtBox), page_rect);
+		break;
+	case FZ_UNKNOWN_BOX:
+		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "unknown page box type: %d", box);
+	}
 }
 
 fz_link *
@@ -638,16 +688,14 @@ void
 pdf_page_obj_transform_box(fz_context *ctx, pdf_obj *pageobj, fz_rect *outbox, fz_matrix *page_ctm, fz_box_type box)
 {
 	pdf_obj *obj;
-	fz_rect usedbox, tempbox, mediabox;
+	fz_rect usedbox, tempbox, cropbox;
 	float userunit = 1;
 	int rotate;
 
 	if (!outbox)
 		outbox = &tempbox;
 
-	obj = pdf_dict_get(ctx, pageobj, PDF_NAME(UserUnit));
-	if (pdf_is_number(ctx, obj))
-		userunit = pdf_to_real(ctx, obj);
+	userunit = pdf_dict_get_real_default(ctx, pageobj, PDF_NAME(UserUnit), 1);
 
 	obj = NULL;
 	if (box == FZ_ART_BOX)
@@ -674,7 +722,7 @@ pdf_page_obj_transform_box(fz_context *ctx, pdf_obj *pageobj, fz_rect *outbox, f
 	*outbox = usedbox;
 
 	/* Snap page rotation to 0, 90, 180 or 270 */
-	rotate = pdf_to_int(ctx, pdf_dict_get_inheritable(ctx, pageobj, PDF_NAME(Rotate)));
+	rotate = pdf_dict_get_inheritable_int(ctx, pageobj, PDF_NAME(Rotate));
 	if (rotate < 0)
 		rotate = 360 - ((-rotate) % 360);
 	if (rotate >= 360)
@@ -692,38 +740,41 @@ pdf_page_obj_transform_box(fz_context *ctx, pdf_obj *pageobj, fz_rect *outbox, f
 	/* Rotate */
 	*page_ctm = fz_pre_rotate(*page_ctm, -rotate);
 
-	/* Always use MediaBox to set origin to top left */
-	mediabox = pdf_to_rect(ctx, pdf_dict_get_inheritable(ctx, pageobj, PDF_NAME(MediaBox)));
-	if (fz_is_empty_rect(mediabox))
-		mediabox = fz_make_rect(0, 0, 612, 792);
-	mediabox.x0 = fz_min(mediabox.x0, mediabox.x1);
-	mediabox.y0 = fz_min(mediabox.y0, mediabox.y1);
-	mediabox.x1 = fz_max(mediabox.x0, mediabox.x1);
-	mediabox.y1 = fz_max(mediabox.y0, mediabox.y1);
-	if (mediabox.x1 - mediabox.x0 < 1 || mediabox.y1 - mediabox.y0 < 1)
-		mediabox = fz_unit_rect;
+	/* Always use CropBox to set origin to top left */
+	obj = pdf_dict_get_inheritable(ctx, pageobj, PDF_NAME(CropBox));
+	if (!pdf_is_array(ctx, obj))
+		obj = pdf_dict_get_inheritable(ctx, pageobj, PDF_NAME(MediaBox));
+	cropbox = pdf_to_rect(ctx, obj);
+	if (fz_is_empty_rect(cropbox))
+		cropbox = fz_make_rect(0, 0, 612, 792);
+	cropbox.x0 = fz_min(cropbox.x0, cropbox.x1);
+	cropbox.y0 = fz_min(cropbox.y0, cropbox.y1);
+	cropbox.x1 = fz_max(cropbox.x0, cropbox.x1);
+	cropbox.y1 = fz_max(cropbox.y0, cropbox.y1);
+	if (cropbox.x1 - cropbox.x0 < 1 || cropbox.y1 - cropbox.y0 < 1)
+		cropbox = fz_unit_rect;
 
-	/* Translate page origin of MediaBox to 0,0 */
-	mediabox = fz_transform_rect(mediabox, *page_ctm);
-	*page_ctm = fz_concat(*page_ctm, fz_translate(-mediabox.x0, -mediabox.y0));
+	/* Translate page origin of CropBox to 0,0 */
+	cropbox = fz_transform_rect(cropbox, *page_ctm);
+	*page_ctm = fz_concat(*page_ctm, fz_translate(-cropbox.x0, -cropbox.y0));
 }
 
 void
-pdf_page_obj_transform(fz_context *ctx, pdf_obj *pageobj, fz_rect *page_mediabox, fz_matrix *page_ctm)
+pdf_page_obj_transform(fz_context *ctx, pdf_obj *pageobj, fz_rect *page_cropbox, fz_matrix *page_ctm)
 {
-	pdf_page_obj_transform_box(ctx, pageobj, page_mediabox, page_ctm, FZ_MEDIA_BOX);
+	pdf_page_obj_transform_box(ctx, pageobj, page_cropbox, page_ctm, FZ_CROP_BOX);
 }
 
 void
-pdf_page_transform_box(fz_context *ctx, pdf_page *page, fz_rect *page_mediabox, fz_matrix *page_ctm, fz_box_type box)
+pdf_page_transform_box(fz_context *ctx, pdf_page *page, fz_rect *page_cropbox, fz_matrix *page_ctm, fz_box_type box)
 {
-	pdf_page_obj_transform_box(ctx, page->obj, page_mediabox, page_ctm, box);
+	pdf_page_obj_transform_box(ctx, page->obj, page_cropbox, page_ctm, box);
 }
 
 void
-pdf_page_transform(fz_context *ctx, pdf_page *page, fz_rect *mediabox, fz_matrix *ctm)
+pdf_page_transform(fz_context *ctx, pdf_page *page, fz_rect *cropbox, fz_matrix *ctm)
 {
-	pdf_page_transform_box(ctx, page, mediabox, ctm, FZ_MEDIA_BOX);
+	pdf_page_transform_box(ctx, page, cropbox, ctm, FZ_CROP_BOX);
 }
 
 static void
@@ -743,7 +794,7 @@ find_seps(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *c
 	if (pdf_name_eq(ctx, nameobj, PDF_NAME(Separation)))
 	{
 		fz_colorspace *cs;
-		const char *name = pdf_to_name(ctx, pdf_array_get(ctx, obj, 1));
+		const char *name = pdf_array_get_name(ctx, obj, 1);
 
 		/* Skip 'special' colorants. */
 		if (!strcmp(name, "Black") ||
@@ -766,6 +817,8 @@ find_seps(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *c
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			return; /* ignore broken colorspace */
 		}
 		fz_try(ctx)
@@ -816,7 +869,7 @@ find_devn(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *c
 	for (j = 0; j < m; j++)
 	{
 		fz_colorspace *cs;
-		const char *name = pdf_to_name(ctx, pdf_array_get(ctx, arr, j));
+		const char *name = pdf_array_get_name(ctx, arr, j);
 
 		/* Skip 'special' colorants. */
 		if (!strcmp(name, "Black") ||
@@ -841,6 +894,8 @@ find_devn(fz_context *ctx, fz_separations **seps, pdf_obj *obj, pdf_mark_list *c
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				continue; /* ignore broken colorspace */
 			}
 			fz_try(ctx)
@@ -1010,7 +1065,11 @@ pdf_load_default_colorspaces_imp(fz_context *ctx, fz_default_colorspaces *defaul
 			fz_drop_colorspace(ctx, cs);
 		}
 		fz_catch(ctx)
+		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
+		}
 	}
 
 	cs_obj = pdf_dict_get(ctx, obj, PDF_NAME(DefaultRGB));
@@ -1023,7 +1082,11 @@ pdf_load_default_colorspaces_imp(fz_context *ctx, fz_default_colorspaces *defaul
 			fz_drop_colorspace(ctx, cs);
 		}
 		fz_catch(ctx)
+		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
+		}
 	}
 
 	cs_obj = pdf_dict_get(ctx, obj, PDF_NAME(DefaultCMYK));
@@ -1036,7 +1099,11 @@ pdf_load_default_colorspaces_imp(fz_context *ctx, fz_default_colorspaces *defaul
 			fz_drop_colorspace(ctx, cs);
 		}
 		fz_catch(ctx)
+		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
+		}
 	}
 }
 
@@ -1068,6 +1135,7 @@ pdf_load_default_colorspaces(fz_context *ctx, pdf_document *doc, pdf_page *page)
 			fz_drop_default_colorspaces(ctx, default_cs);
 			fz_rethrow(ctx);
 		}
+		fz_ignore_error(ctx);
 		page->super.incomplete = 1;
 	}
 
@@ -1116,6 +1184,15 @@ pdf_load_page_imp(fz_context *ctx, fz_document *doc_, int chapter, int number)
 	pdf_annot *annot;
 	pdf_obj *pageobj, *obj;
 
+	if (doc->is_fdf)
+		fz_throw(ctx, FZ_ERROR_FORMAT, "FDF documents have no pages");
+
+	if (chapter != 0)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "invalid chapter number: %d", chapter);
+
+	if (number < 0 || number >= pdf_count_pages(ctx, doc))
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "invalid page number: %d", number);
+
 	if (doc->file_reading_linearly)
 	{
 		pageobj = pdf_progressive_advance(ctx, doc, number);
@@ -1134,9 +1211,9 @@ pdf_load_page_imp(fz_context *ctx, fz_document *doc_, int chapter, int number)
 		obj = pdf_dict_get(ctx, pageobj, PDF_NAME(Annots));
 		if (obj)
 		{
-			fz_rect page_mediabox;
+			fz_rect page_cropbox;
 			fz_matrix page_ctm;
-			pdf_page_transform(ctx, page, &page_mediabox, &page_ctm);
+			pdf_page_transform(ctx, page, &page_cropbox, &page_ctm);
 			page->links = pdf_load_link_annots(ctx, doc, page, obj, number, page_ctm);
 			pdf_load_annots(ctx, page, obj);
 		}
@@ -1148,6 +1225,7 @@ pdf_load_page_imp(fz_context *ctx, fz_document *doc_, int chapter, int number)
 			fz_drop_page(ctx, &page->super);
 			fz_rethrow(ctx);
 		}
+		fz_ignore_error(ctx);
 		page->super.incomplete = 1;
 		fz_drop_link(ctx, page->links);
 		page->links = NULL;
@@ -1213,6 +1291,7 @@ pdf_load_page_imp(fz_context *ctx, fz_document *doc_, int chapter, int number)
 			fz_drop_page(ctx, &page->super);
 			fz_rethrow(ctx);
 		}
+		fz_ignore_error(ctx);
 		page->super.incomplete = 1;
 	}
 
@@ -1324,8 +1403,8 @@ pdf_add_page(fz_context *ctx, pdf_document *doc, fz_rect mediabox, int rotate, p
 	}
 	fz_catch(ctx)
 	{
-		pdf_abandon_operation(ctx, doc);
 		pdf_drop_obj(ctx, page_obj);
+		pdf_abandon_operation(ctx, doc);
 		fz_rethrow(ctx);
 	}
 	return page_ref;
@@ -1343,7 +1422,7 @@ pdf_insert_page(fz_context *ctx, pdf_document *doc, int at, pdf_obj *page_ref)
 	if (at == INT_MAX)
 		at = count;
 	if (at > count)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot insert page beyond end of page tree");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot insert page beyond end of page tree");
 
 	pdf_begin_operation(ctx, doc, "Insert page");
 
@@ -1354,10 +1433,10 @@ pdf_insert_page(fz_context *ctx, pdf_document *doc, int at, pdf_obj *page_ref)
 			pdf_obj *root = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root));
 			parent = pdf_dict_get(ctx, root, PDF_NAME(Pages));
 			if (!parent)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find page tree");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "cannot find page tree");
 			kids = pdf_dict_get(ctx, parent, PDF_NAME(Kids));
 			if (!kids)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "malformed page tree");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "malformed page tree");
 			pdf_array_insert(ctx, kids, page_ref, 0);
 		}
 		else if (at == count)
@@ -1609,8 +1688,7 @@ pdf_adjust_page_labels(fz_context *ctx, pdf_document *doc, int index, int adjust
 
 		// Increase/decrease the indices in the name tree
 		for (; i < pdf_array_len(ctx, range.nums); i += 2)
-			pdf_array_put_drop(ctx, range.nums, i,
-				pdf_new_int(ctx, pdf_array_get_int(ctx, range.nums, i) + adjust));
+			pdf_array_put_int(ctx, range.nums, i, pdf_array_get_int(ctx, range.nums, i) + adjust);
 
 		// TODO: delete page labels that have no effect (zero range)
 
@@ -1780,3 +1858,15 @@ pdf_page_label_imp(fz_context *ctx, fz_document *doc, int chapter, int page, cha
 {
 	pdf_page_label(ctx, pdf_document_from_fz_document(ctx, doc), page, buf, size);
 }
+
+pdf_page *
+pdf_keep_page(fz_context *ctx, pdf_page *page)
+{
+	return (pdf_page *) fz_keep_page(ctx, &page->super);
+}
+
+void
+pdf_drop_page(fz_context *ctx, pdf_page *page)
+{
+	fz_drop_page(ctx, &page->super);
+}
